#This simulation investigation intends to observe the behavior of a number of TCP flows
#when traversing through switches of different bandwidths and different buffer sizes.
#
#The question is:  is TCP performance better when going through a slower switch, but with
#larger buffers, then through a higher-speed switch with small buffers, or is TCP
#performance better when going through a higher speed switch only with small buffers?
#
#In this simulation, we will use 10 Gbps switches only.  Buffer sizes of
#each will be modified.
#
#Observe an important comment about ns behavior regarding the TCP variables cwnd_ and
#window_, in the TCP source creation section further below.
#
# August 2010 Marcos Portnoi.

# --- Random Numbers and Seed
#Seed default random number generator with current time
global defaultRNG
$defaultRNG seed 0

#Number of TCP traffic sources (used to create traffic generators, nodes, etc.)
set tcpGenerators 4

#Turn tracing on=1 or off=0
#be careful:  ns can generate enormous trace files
set traceAll 0
set traceQueue 1

#Queue size limits; ns default is 50
set tcpSourceQueueLimit 50
set tcpConcurTrafficQueueLimit 50
set switch10GbpsQueueLimit 50
#set switch1GbpsQueueLimit 1000

#Create simulator object
set ns [new Simulator]
set MAX_TIME 3000.0; #set maximum simulation time; use at least a decimal to prevent this constant being interpreted as a integer

# --- Define color index (class color)
# list of colors available at ~ns/man/mann/colors.n
$ns color 0 red
$ns color 1 blue
$ns color 2 chocolate
$ns color 3 yellow
$ns color 4 green
$ns color 5 tan
$ns color 6 gold
$ns color 7 black
$ns color 8 white
$ns color 9 darkblue
$ns color 10 cyan
$ns color 11 magenta
$ns color 12 orange
$ns color 13 bisque
$ns color 14 purple
$ns color 15 coral
$ns color 16 grey
$ns color 17 khaki
$ns color 18 AntiqueWhite
$ns color 19 aquamarine
$ns color 20 azure
$ns color 21 DarkSalmon
$ns color 22 DarkSeaGreen
$ns color 23 firebrick

#Use dinamic routing
#$ns rtproto DV

#Open files for trace if corresponding flags are set
if {$traceAll} {
	set nf [open out.nam w]
	set tr [open out.tr w]
	$ns namtrace-all $nf
	$ns trace-all $tr
}
if {$traceQueue} {
	set queueStats [open queueStats.csv w]
	set queueAvgStats [open queueAvgStats.csv w]
}

set throughputs [open throughputs.csv w]
set tcpVariablesTrace [open tcpVariablesTrace.csv w]
set simParams [open simParams.txt w]

#=========================================================================
#Create nodes
#create main TCP sources and sinks
for {set i 0} {$i < $tcpGenerators} {incr i} {
	set source($i) [$ns node]
	set sink($i) [$ns node]
}

#set switch1Gbps [$ns node]
set switch10Gbps [$ns node]
set switch10GbpsOut [$ns node]
#Concurrent traffic
set concurTraffic [$ns node]
set sinkConcurTraffic [$ns node]

#Connect nodes
#default queue size limit is 50
for {set i 0} {$i < $tcpGenerators} {incr i} {
	$ns duplex-link $source($i) $switch10Gbps 10000Mb 1ms DropTail
	$ns queue-limit $source($i) $switch10Gbps $tcpSourceQueueLimit
	$ns duplex-link $switch10GbpsOut $sink($i) 10000Mb 1ms DropTail
	$ns queue-limit $switch10GbpsOut $sink($i) $switch10GbpsQueueLimit
}

#$ns duplex-link $switch1Gbps $switch10Gbps 1000Mb 1ms DropTail
#$ns queue-limit $switch1Gbps $switch10Gbps $switch1GbpsQueueLimit
$ns duplex-link $switch10Gbps $switch10GbpsOut 10000Mb 10ms DropTail
$ns queue-limit $switch10Gbps $switch10GbpsOut $switch10GbpsQueueLimit
$ns duplex-link $concurTraffic $switch10Gbps 10000Mb 10ms DropTail
$ns queue-limit $concurTraffic $switch10Gbps $tcpConcurTrafficQueueLimit
$ns duplex-link $switch10GbpsOut $sinkConcurTraffic 10000Mb 1ms DropTail
$ns queue-limit $switch10GbpsOut $sinkConcurTraffic $switch10GbpsQueueLimit

#Set visual orientation for nam
for {set i 0} {$i < $tcpGenerators} {incr i} {
	$ns duplex-link-op $source($i) $switch10Gbps queuePos 0.5
	$ns duplex-link-op $switch10GbpsOut $sink($i) queuePos 0.5
}
#$ns duplex-link-op $switch1Gbps $switch10Gbps queuePos 0.5
$ns duplex-link-op $switch10Gbps $switch10GbpsOut queuePos 0.5
$ns duplex-link-op $concurTraffic $switch10Gbps queuePos 0.5
$ns duplex-link-op $switch10GbpsOut $sinkConcurTraffic queuePos 0.0

#modify here for automatic placement of n tcp sources
$ns duplex-link-op $source(0) $switch10Gbps orient 1.67
$ns duplex-link-op $source(1) $switch10Gbps orient 1.83
$ns duplex-link-op $source(2) $switch10Gbps orient 0.17
$ns duplex-link-op $source(3) $switch10Gbps orient 0.33
#$ns duplex-link-op $switch1Gbps $switch10Gbps orient 0.0
$ns duplex-link-op $switch10Gbps $switch10GbpsOut orient 0.0
#modify here for automatic placement of n tcp sinks
$ns duplex-link-op $switch10GbpsOut $sink(0) orient 0.33
$ns duplex-link-op $switch10GbpsOut $sink(1) orient 0.17
$ns duplex-link-op $switch10GbpsOut $sink(2) orient 1.83
$ns duplex-link-op $switch10GbpsOut $sink(3) orient 1.67
$ns duplex-link-op $concurTraffic $switch10Gbps orient 1.5
$ns duplex-link-op $switch10GbpsOut $sinkConcurTraffic orient 1.5

#=========================================================================================================
#Create TCP agents and TCPSinks, attach them to nodes and link them
#
#Also, set here the maximum window_ size (cwnd_ or congestion window) for TCP
#Here is the catch:  ns allows TCP algorithm to grow cnwd accordingly if there is no congestion.  The
#TCP variables trace file can demonstrate that.  However, the maximum burst of TCP PDUs (segments) is
#limited to the maximum (advertised) window_ size, regardless whether cwnd_ is larger.  Thus, if in a specific moment,
#cwnd_ value is 2000 and window_ is set to 1000, TCP will send a burst of only 1000 PDUs and then wait for
#the corresponding ACKs (since the max advertised window_ size is 1000).
#
for {set i 0} {$i < $tcpGenerators} {incr i} {
	set connection_list [$ns create-connection-list TCP/Reno $source($i) TCPSink $sink($i) $i]
	set tcpSource($i) [lindex $connection_list 0]
	set tcpSink($i) [lindex $connection_list 1]
	$tcpSource($i) set fid_ $i
	$tcpSink($i) set fid_ $i
	$tcpSource($i) set window_ 64000
	#$tcpSource($i) set windowInit_ 64000
	#$tcpSource($i) set packetSize_ 8000
	$tcpSource($i) set maxburst_ 64000
	$tcpSink($i) set window_ 64000
}
#flowID for Concurrent traffic
set concurId 100
set connection_list [$ns create-connection-list TCP/Reno $concurTraffic TCPSink $sinkConcurTraffic $concurId]
set tcpConcurTraffic [lindex $connection_list 0]
set tcpSinkConcurTraffic [lindex $connection_list 1]
$tcpConcurTraffic set fid_ $concurId
$tcpSinkConcurTraffic set fid_ $concurId
$tcpConcurTraffic set window_ 64000
#$tcpConcurTraffic set windowInit_ 64000
#$tcpConcurTraffic set packetSize_ 8000
$tcpConcurTraffic set maxburst_ 64000
$tcpSinkConcurTraffic set window_ 64000

#Create FTP applications and attach them to agents
for {set i 0} {$i < $tcpGenerators} {incr i} {
	set ftp($i) [new Application/FTP]
	$ftp($i) attach-agent $tcpSource($i)
}
set ftpConcur [new Application/FTP]
$ftpConcur attach-agent $tcpConcurTraffic


#Create queue monitors if corresponding trace flag is set
if {$traceQueue} {
	for {set i 0} {$i < $tcpGenerators} {incr i} {
		set qmonSource($i) [$ns monitor-queue $source($i) $switch10Gbps ""]
		set qmonSink($i) [$ns monitor-queue $switch10GbpsOut $sink($i) ""]
	}
	#set qmonSwitch1_10 [$ns monitor-queue $switch1Gbps $switch10Gbps ""]
	set qmonSwitch10_out [$ns monitor-queue $switch10Gbps $switch10GbpsOut ""]
	set qmonConcur_Switch10 [$ns monitor-queue $concurTraffic $switch10Gbps ""]
	set qmonSwitch10_SinkConcur [$ns monitor-queue $switch10GbpsOut $sinkConcurTraffic ""]
}

#Create statistics record procedure
proc record {time} {
	global tcpSource tcpConcurTraffic tcpSink tcpSinkConcurTraffic throughputs tcpGenerators tcpVariablesTrace
	#Get instance of simulator
	set ns [Simulator instance]
	
	#Get current time
	set now [$ns now]
	puts "Recording at $now s..."
	
	#print header only once to sink throughputs file
	if {$now == 0} {
		puts $throughputs "Throughput MBps"
		puts -nonewline $throughputs "time;"
		#tcpSources
		for {set i 0} {$i < $tcpGenerators} {incr i} {
			puts -nonewline $throughputs "tcpSink($i);"
		}
		#concurrent traffic sink
		puts $throughputs "tcpSinkConcurTraffic"
	}
		
	#How many bytes have been received by traffic sinks?
	# Problem here is that bytes_ in tcp-sink.cc is a INT value... max is 2147483647, and this is easily surpassed in gigabit links; bytes_ then WRAPS
	# To avoid this for 10Gbps of throughput, we can use the relation:
	# time = max_int_bytes_ / throughput
	# where time is the time step or resolution where bytes_ should be read from TCPSink, and then set to zero again.
	# Thus, for the usual max_int, we have:
	# time = 2147483647 / 10,000,000,000 * 8 (multiply by 8 to equalize the units to bytes) = 1.7179 s
	# Anything below that value should work accordingly, without integer wrapping.
	for {set i 0} {$i < $tcpGenerators} {incr i} {
		set bw($i) [$tcpSink($i) set bytes_]
	}
	set bwConcur [$tcpSinkConcurTraffic set bytes_]
		
	#Calculate bandwidth (MBit/s) and write to file
	puts -nonewline $throughputs "$now;"
	puts -nonewline "$now "
	for {set i 0} {$i < $tcpGenerators} {incr i} {
		puts -nonewline $throughputs "[expr $bw($i)/$time*8.0/1000000.0];"
		puts -nonewline "[expr $bw($i)/$time*8.0/1000000.0];"
	}
	#concurrent traffic bandwidth (MBit/s)
	puts $throughputs "[expr $bwConcur/$time*8.0/1000000.0]"
	puts "[expr $bwConcur/$time*8.0/1000000.0]"
	
	#Reset bytes_ values on traffic sinks
	for {set i 0} {$i < $tcpGenerators} {incr i} {
		$tcpSink($i) set bytes_ 0
	}
	$tcpSinkConcurTraffic set bytes_ 0
	
	#record TCP variables
	#print header only once to tcpVariablesTrace file
	if {$now == 0} {
		puts $tcpVariablesTrace "node;time;awnd_;rttvar_;backoff_;cwnd_;ssthresh_;rtt_;srtt_"
	}
	#tcpSources
	for {set i 0} {$i < $tcpGenerators} {incr i} {
		puts $tcpVariablesTrace "tcpSource($i);$now;[$tcpSource($i) set awnd_];[$tcpSource($i) set rttvar_];[$tcpSource($i) set backoff_];[$tcpSource($i) set cwnd_];[$tcpSource($i) set ssthresh_];[$tcpSource($i) set rtt_];[$tcpSource($i) set srtt_]"
	}
	#tcp concurrent traffic
	puts $tcpVariablesTrace "tcpConcurTraffic;$now;[$tcpConcurTraffic set awnd_];[$tcpConcurTraffic set rttvar_];[$tcpConcurTraffic set backoff_];[$tcpConcurTraffic set cwnd_];[$tcpConcurTraffic set ssthresh_];[$tcpConcurTraffic set rtt_];[$tcpConcurTraffic set srtt_]"
	
	#Re-schedule procedure
	$ns at [expr $now+$time] "record $time"
}

#Schedule
$ns at 0.0 "record 1.5"
$ns at 0.0 "puts \"Starting simulation...\""
#ftp generators start synchronized!
for {set i 0} {$i < $tcpGenerators} {incr i} {
	#$ns at [expr 0.0 + $i] "$ftp($i) start"
	$ns at 500.0 "$ftp($i) start"
}
$ns at 0.0 "$ftpConcur start"
for {set i 0} {$i < $tcpGenerators} {incr i} {
	$ns at $MAX_TIME "$ftp($i) stop"
}
$ns at $MAX_TIME "$ftpConcur stop"
$ns at $MAX_TIME "finish"

#Finish procedure
proc finish {} {
	global ns nf tr throughputs tcpGenerators queueStats queueAvgStats tcpSource tcpConcurTraffic simParams
	global qmonSource qmonSink
	#global qmonSwitch1_10
	global qmonSwitch10_out qmonConcur_Switch10 qmonSwitch10_SinkConcur
	global tcpSourceQueueLimit tcpConcurTrafficQueueLimit
	#global switch1GbpsQueueLimit
	global switch10GbpsQueueLimit
	global MAX_TIME traceAll traceQueue
	if {$traceAll} {
		$ns flush-trace
		#close trace files
		close $nf
		close $tr
	}
	close $throughputs
	puts "Ending simulation.  Simulation time:  $MAX_TIME s"
	puts "                                      ====="	
		
	#---------------- generating queue statistics -----------------------
	
	#outputting queue stats to screen, if trace flag is set
	#queue sums stats
	if {$traceQueue} {
		puts "\nStatistics:"  
		puts "                                 Arrived                              Lost                      Departed"
		puts "                    Packets      Bytes    Throughput(Mbps)    Packets     Bytes    Packets     Bytes    Throughput(Mbps)"
		puts "-----------------------------------------------------------------------------------"
		
		#sources
		for {set i 0} {$i < $tcpGenerators} {incr i} {
			puts -nonewline [format "Queue qmonSource($i):   %7d   %10d   %10f" [eval $qmonSource($i) set parrivals_] [eval $qmonSource($i) set barrivals_] [expr [$qmonSource($i) set barrivals_]/$MAX_TIME*8.0/1000000.0]]
			puts -nonewline [format "  %7d   %10f" [eval $qmonSource($i) set pdrops_] [eval $qmonSource($i) set bdrops_]]
			puts [format "  %7d   %10d   %10f" [eval $qmonSource($i) set pdepartures_] [eval $qmonSource($i) set bdepartures_] [expr [$qmonSource($i) set bdepartures_]/$MAX_TIME*8.0/1000000.0]]
		}
		#sinks
		for {set i 0} {$i < $tcpGenerators} {incr i} {
			puts -nonewline [format "Queue qmonSink($i):   %7d   %10d   %10f" [eval $qmonSink($i) set parrivals_] [eval $qmonSink($i) set barrivals_] [expr [$qmonSink($i) set barrivals_]/$MAX_TIME*8.0/1000000.0]]
			puts -nonewline [format "  %7d   %10d" [eval $qmonSink($i) set pdrops_] [eval $qmonSink($i) set bdrops_]]
			puts [format "  %7d   %10d   %10f" [eval $qmonSink($i) set pdepartures_] [eval $qmonSink($i) set bdepartures_] [expr [$qmonSink($i) set bdepartures_]/$MAX_TIME*8.0/1000000.0]]
		}
		#others
		set j 0
		foreach i {$qmonSwitch10_out $qmonConcur_Switch10 $qmonSwitch10_SinkConcur} {
			puts -nonewline [format "Queue $i:   %7d   %10d   %10f" [eval $i set parrivals_] [eval $i set barrivals_] [expr [eval $i set barrivals_]/$MAX_TIME*8.0/1000000.0]]
			puts -nonewline [format "  %7d   %10d" [eval $i set pdrops_] [eval $i set bdrops_]]
			puts [format "  %7d   %10d   %10f" [eval $i set pdepartures_] [eval $i set bdepartures_] [expr [eval $i set bdepartures_]/$MAX_TIME*8.0/1000000.0]]
		}
		
		#average queue sizes
		puts "\nAverage Queue Size:        Packets         Bytes    sum_ (pkts)"
		puts "-----------------------------------------------------------------------"
		
		#sources
		for {set i 0} {$i < $tcpGenerators} {incr i} {
			set bytesInt($i) [eval $qmonSource($i) get-bytes-integrator]
			set pktsInt($i) [eval $qmonSource($i) get-pkts-integrator]
			set avg_queue_b($i) [expr [$bytesInt($i) set sum_]/$MAX_TIME]
			set avg_queue_p($i) [expr [$pktsInt($i) set sum_]/$MAX_TIME]
			puts [format "Queue Source($i):          %5.2f      %8.2f    [$pktsInt($i) set sum_]" $avg_queue_p($i) $avg_queue_b($i)]
		}
			
			
		#sinks
		for {set i 0} {$i < $tcpGenerators} {incr i} {
			set bytesInt($i) [eval $qmonSink($i) get-bytes-integrator]
			set pktsInt($i) [eval $qmonSink($i) get-pkts-integrator]
			set avg_queue_b($i) [expr [$bytesInt($i) set sum_]/$MAX_TIME]
			set avg_queue_p($i) [expr [$pktsInt($i) set sum_]/$MAX_TIME]
			puts [format "Queue Sink($i):          %5.2f      %8.2f    [$pktsInt($i) set sum_]" $avg_queue_p($i) $avg_queue_b($i)]
		}
		
		#others
		set j 0
		foreach i {$qmonSwitch10_out $qmonConcur_Switch10 $qmonSwitch10_SinkConcur} {
			incr j
			set bytesInt($j) [eval $i get-bytes-integrator]
			set pktsInt($j) [eval $i get-pkts-integrator]
			set avg_queue_b($j) [expr [$bytesInt($j) set sum_]/$MAX_TIME]
			set avg_queue_p($j) [expr [$pktsInt($j) set sum_]/$MAX_TIME]
			puts [format "Queue $i:          %5.2f      %8.2f    [$pktsInt($j) set sum_]" $avg_queue_p($j) $avg_queue_b($j)]
		}
		#-----------------------
		#outputting to csv files
		#queue sums stats
		puts $queueStats "Simulation time;$MAX_TIME"
		puts $queueStats "node;arrived_packets;arrived_bytes;arrived_throughputMbps;lost_packets;lost_bytes;departed_packets;departed_bytes;departed_throughputMbps"
		
		#sources
		for {set i 0} {$i < $tcpGenerators} {incr i} {
			puts $queueStats "qmonSource($i);[eval $qmonSource($i) set parrivals_];[eval $qmonSource($i) set barrivals_];[expr [$qmonSource($i) set barrivals_]/$MAX_TIME*8.0/1000000.0];[eval $qmonSource($i) set pdrops_];[eval $qmonSource($i) set bdrops_];[eval $qmonSource($i) set pdepartures_];[eval $qmonSource($i) set bdepartures_];[expr [$qmonSource($i) set bdepartures_]/$MAX_TIME*8.0/1000000.0]"	
		}
		#sinks
		for {set i 0} {$i < $tcpGenerators} {incr i} {
			puts $queueStats "qmonSink($i);[eval $qmonSink($i) set parrivals_];[eval $qmonSink($i) set barrivals_];[expr [$qmonSink($i) set barrivals_]/$MAX_TIME*8.0/1000000.0];[eval $qmonSink($i) set pdrops_];[eval $qmonSink($i) set bdrops_];[eval $qmonSink($i) set pdepartures_];[eval $qmonSink($i) set bdepartures_];[expr [$qmonSink($i) set bdepartures_]/$MAX_TIME*8.0/1000000.0]"	
		}
		#others
		set j 0
		foreach i {$qmonSwitch10_out $qmonConcur_Switch10 $qmonSwitch10_SinkConcur} {
			puts $queueStats "$i;[eval $i set parrivals_];[eval $i set barrivals_];[expr [eval $i set barrivals_]/$MAX_TIME*8.0/1000000.0];[eval $i set pdrops_];[eval $i set bdrops_];[eval $i set pdepartures_];[eval $i set bdepartures_];[expr [eval $i set bdepartures_]/$MAX_TIME*8.0/1000000.0]"	
		}
		
		#average queue sizes
		puts $queueAvgStats "queue;average_queue_size_packets;average_queue_size_bytes;sum_pkts"
		#sources
		for {set i 0} {$i < $tcpGenerators} {incr i} {
			set bytesInt($i) [eval $qmonSource($i) get-bytes-integrator]
			set pktsInt($i) [eval $qmonSource($i) get-pkts-integrator]
			set avg_queue_b($i) [expr [$bytesInt($i) set sum_]/$MAX_TIME]
			set avg_queue_p($i) [expr [$pktsInt($i) set sum_]/$MAX_TIME]
			puts $queueAvgStats "Source($i);$avg_queue_p($i);$avg_queue_b($i);[$pktsInt($i) set sum_]"
		}
			
			
		#sinks
		for {set i 0} {$i < $tcpGenerators} {incr i} {
			set bytesInt($i) [eval $qmonSink($i) get-bytes-integrator]
			set pktsInt($i) [eval $qmonSink($i) get-pkts-integrator]
			set avg_queue_b($i) [expr [$bytesInt($i) set sum_]/$MAX_TIME]
			set avg_queue_p($i) [expr [$pktsInt($i) set sum_]/$MAX_TIME]
			puts $queueAvgStats "Sink($i);$avg_queue_p($i);$avg_queue_b($i);[$pktsInt($i) set sum_]"
		}
		
		#others
		set j 0
		foreach i {$qmonSwitch10_out $qmonConcur_Switch10 $qmonSwitch10_SinkConcur} {
			incr j
			set bytesInt($j) [eval $i get-bytes-integrator]
			set pktsInt($j) [eval $i get-pkts-integrator]
			set avg_queue_b($j) [expr [$bytesInt($j) set sum_]/$MAX_TIME]
			set avg_queue_p($j) [expr [$pktsInt($j) set sum_]/$MAX_TIME]
			puts $queueAvgStats "$i;$avg_queue_p($j);$avg_queue_b($j);[$pktsInt($j) set sum_]"
		}
		#--------------------
	}
	#------------ end of queue statistics generation -------------------
	
	#------ Print simulation parameters -------
	
	#print simulation time
	puts $simParams "Simulation time:  $MAX_TIME s\n"
	
	#print queue limits
	puts $simParams "Queue Limits:"
	puts $simParams "Queue                          Limit (packets)"
	puts $simParams "----------------------------------------------"
	foreach i {$tcpSourceQueueLimit $tcpConcurTrafficQueueLimit $switch10GbpsQueueLimit} {
		puts $simParams [format "%-28s:     %5d" $i [expr $i]]
	}
	
	#print TCP sources configuration parameters
	puts $simParams "\n\nTCP Configuration Parameters:"
	puts $simParams "node               window_    windowInit_   packetSize_    tcpTick_   maxburst_    maxcwnd_"
	puts $simParams "-------------------------------------------------------------------------------------------"
	for {set i 0} {$i < $tcpGenerators} {incr i} {
		puts $simParams [format "tcpSource($i):        %5d          %5d         %5d       %5.3f      %6d      %6d" [$tcpSource($i) set window_] [$tcpSource($i) set windowInit_] [$tcpSource($i) set packetSize_] [$tcpSource($i) set tcpTick_] [$tcpSource($i) set maxburst_] [$tcpSource($i) set maxcwnd_]]
	}
	puts $simParams [format "tcpConcurTraffic:   %6d          %5d         %5d       %5.3f      %6d      %6d" [$tcpConcurTraffic set window_] [$tcpConcurTraffic set windowInit_] [$tcpConcurTraffic set packetSize_] [$tcpConcurTraffic set tcpTick_] [$tcpConcurTraffic set maxburst_] [$tcpConcurTraffic set maxcwnd_]]
	
	#------ end of printing simulation parameters ------ 
	
	if {$traceAll} {
		puts "Generating nam..."
		exec nam out.nam &
		#puts "Generating xgraph..."
		#exec xgraph throughputs.csv -geometry 800x600 &
	}
	exit 0
}

$ns run